/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/**********************************************************************
grid_proxy_init.c

Description:
	Program used to get a session "proxy" certificate using
	your long term certificate. This is functionally equivelent 
	to the Kerberos kinit program. 

CVS Information:

	$Source: /cvs/nordugrid/grid-manager/external/voms-1.1.39/client/Attic/voms_proxy_init.cc,v $
	$Date: 2003/11/21 16:59:27 $
	$Revision: 1.4 $
	$Author: aleks $

**********************************************************************/

static char *rcsid = "$Header: /cvs/nordugrid/grid-manager/external/voms-1.1.39/client/Attic/voms_proxy_init.cc,v 1.4 2003/11/21 16:59:27 aleks Exp $";

/**********************************************************************
                             Include header files
**********************************************************************/
#include "config.h"
#include <string>

const std::string SUBPACKAGE      = "voms-proxy-init";
//const std::string VERSION         = "0.1";
const std::string CONFILENAME     = "/opt/edg/etc/vomses";
const std::string USERCONFILENAME = ".edg/vomses";

extern "C" {

#include <gssapi.h>
// #include "sslutils.h"
#include <globus_gsi_credential.h>
#include <globus_gsi_system_config.h>
#include <globus_gsi_proxy.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#include "openssl/buffer.h"
#include "openssl/crypto.h"
#include "openssl/objects.h"
#include "openssl/asn1.h"
#include "openssl/evp.h"
#include "openssl/x509.h"
#include "openssl/pem.h"
#include "openssl/ssl.h"
#include "openssl/rsa.h"

#ifdef USE_PKCS11
#include "scutils.h"
#endif

// extern int proxy_verify_cert_chain(X509 * ucert, STACK_OF(X509) * cert_chain,
//				   proxy_verify_desc * pvd);

//extern void proxy_verify_ctx_init(proxy_verify_ctx_desc * pvxd);
}


#include "data.h"
#include "client.h"
#include "options.h"
#include "file.h"

#include <vector>
#include <iostream>
#include <fstream>
#include <iterator>
#include <algorithm>

int fix_add_entry_asn1_set_param;
bool debug = false;
bool quiet = false;

/**********************************************************************
                       Define module specific variables
**********************************************************************/

extern "C" {
static int pwstdin_callback(char *, int, int);
static void kpcallback(int, int, void*);
}
static bool make_proxy();
static int pclocal(globus_gsi_cred_handle_t, const char *, int, int, int, int, int, void (*)(int,int,void*), std::string &, char *, int, char *, int);
static X509_EXTENSION *proxy_extension_voms_create(std::string &);
static X509_EXTENSION *proxy_extension_vo_create(std::string &);
static X509_EXTENSION *proxy_extension_file_create(unsigned char *, size_t);



#ifdef WIN32
static int getuid() { return 0;}
#endif
/**********************************************************************
Function: pwstdin_callback()

Description:
	Get the pass-phrase frm stdin. Used by some scripts 
	or graphical interfaces rather then the prompt to the 
	terminal. See SSLeay src/crypto/pem/pem_lib.c for 
	the def_callback(). 

Parameters:
	buf location to store the pass-phrase
	num length of the buf
	w   0 only need to prompt once, 1 verify by asking twice. 
		w is not needed here. 
	

Returns:
	length of pass-phrase returned
	-1 on error
**********************************************************************/

extern "C" {
static int
pwstdin_callback(char * buf, int num, int w)
{
  int i;

  if (!(fgets(buf, num, stdin))) {
    std::cerr << "Failed to read pass-phrase from stdin" << std::endl;
    return -1;
  }
  i = strlen(buf);
  if (buf[i-1] == '\n') {
    buf[i-1] = '\0';
    i--;
  }
  return i;	

}
}

/**********************************************************************
Function: kpcallback()

Description:
	prints the ...+++ during the key generation 
	Not clear if this assists in randomizing or just to let
	the user know its working

Parameters:

Returns:
**********************************************************************/


#define MAXVOS 2000
static std::vector<char *>voms;
static std::vector<char *>order;
static std::string vo;

static void
kpcallback(int p, int n, void* dummy)
{
    char c='B';

	if (quiet) return;

    if (p == 0) c='.';
    if (p == 1) c='+';
    if (p == 2) c='*';
    if (p == 3) c='\n';
	if (!debug) c = '.';
    fputc(c,stderr);
}

/********************************************************************/

static std::string        certfile;
static std::string        keyfile;
static std::string        outfile;
static std::string        certdir;
static std::string        certcafile;
static std::string        pin         = "";
static std::string        argp;
static std::string        program;
static std::string        conffile    = CONFILENAME;
static std::string        userconf    = "";
static std::string        incfile     = "";
static int                bits        = 512;   /* default proxy to 512 bits */
static int                hours       = 12;    /* default to a 12 hour cert */
static bool               limit_proxy = false; /* dont restrict the proxy */
static bool               independent_proxy = false;
static bool               old_proxy   = false;
static bool               verify      = false;
static bool               print       = false;
static bool               noregen     = false;
static int                version     = 0;
static bool               progversion = false;
static bool               ignorewarn  = false;
static bool               failonwarn  = false;
static int                vomslife    = -1;
static std::string        separate    = "";
#ifdef CLASS_ADD
static void *		  class_add_buf = NULL;
static size_t		  class_add_buf_len = 0;
#endif
extern "C" {
static int (*pw_cb)() = NULL;
};

static std::string             voID = ""; /* Name of the VO, to include in the
					* certificate in a /D=vo field.*/
static bool               novo = false; /* Don't include the /D=vo field in the
					* certificate. */  
static bool new_asn1 = false;
static bool pwstdin = false;

int
main(int argc, char **argv)
{
    struct rlimit newlimit = {0,0};
    
    if (strrchr(argv[0],'/'))
	program = strrchr(argv[0],'/') + 1;
    else
	program = argv[0];
  
    char *vo = getenv("VO");
    if (vo != NULL && strcmp(vo, "") != 0)
      voID=std::string(vo);

    static char *LONG_USAGE = \
      "\n" \
      "    Options\n" \
      "    -help, -usage           Displays usage\n" \
      "    -version                Displays version\n" \
      "    -debug                  Enables extra debug output\n" \
      "    -q                      Quiet mode, minimal output\n" \
      "    -verify                 Verifies certificate to make proxy for\n" \
      "    -pwstdin                Allows passphrase from stdin\n" \
      "    -limited                Creates a limited proxy\n" \
      "    -independent            Creates a independent globus proxy\n" \
      "    -old                    Creates a legacy globus proxy\n" \
      "    -hours H                Proxy is valid for H hours (default:12)\n" \
      "    -bits  B                Number of bits in key {512|1024|2048|4096}\n" \
      "    -cert     <certfile>    Non-standard location of user certificate\n" \
      "    -key      <keyfile>     Non-standard location of user key\n" \
      "    -certdir  <certdir>     Non-standard location of trusted cert dir\n" \
      "    -out      <proxyfile>   Non-standard location of new proxy cert\n" \
      "    -vo <voname>            Specifies a particular VO.\n" \
      "    -novo                   Doesn't specify a particular VO (default).\n" \
      "    -voms <voms<:command>>  Specify voms server. :command is optional.\n" \
      "    --order <group<:role>>  Specify ordering of attributes.\n" \
      "    -vomslife <H>           Try to get a VOMS pseudocert valid for H hours.\n" \
      "    -include <file>         Include the contents of the specified file.\n" \
      "    -conf <file>            Read options from <file>.\n" \
      "    -confile <file>         Non-standard location of voms server addresses.\n" \
      "    -userconf <file>        Non-standard location of user-defined voms server addresses.\n" \
      "    -print                  Print data instead of creating certificates.\n" \
      "    -globus                 Globus version.\n" \
      "    -noregen                Doesn't regenerate a new proxy for the connection.\n"
      "    -separate <file>        Saves the informations returned by the server on file <file>.\n"
      "    -ignorewarn             Ignore warnings.\n"
      "    -failonwarn             Treat warnings as errors.\n"
      "\n";

    set_usage(LONG_USAGE);

    struct option opts[] = {
      {"help",       0, NULL,                OPT_HELP},
      {"usage",      0, NULL,                OPT_HELP},
      {"version",    0, (int *)&progversion, OPT_BOOL},
      {"cert",       1, (int *)&certfile,    OPT_STRING},
      {"certdir",    1, (int *)&certdir,     OPT_STRING},
      {"out",        1, (int *)&outfile,     OPT_STRING},
      {"key",        1, (int *)&keyfile,     OPT_STRING},
      {"include",    1, (int *)&incfile,     OPT_STRING},
      {"hours",      1, &hours,              OPT_NUM},
      {"vomslife",   1, &vomslife,           OPT_NUM},
      {"bits",       1, &bits,               OPT_NUM},
      {"debug",      0, (int *)&debug,       OPT_BOOL},
      {"limited",    0, (int *)&limit_proxy, OPT_BOOL},
      {"independent",0, (int *)&independent_proxy, OPT_BOOL},
      {"old",        0, (int *)&old_proxy,   OPT_BOOL},
      {"verify",     0, (int *)&verify,      OPT_BOOL},
      {"q",          0, (int *)&quiet,       OPT_BOOL},
      {"quiet",      0, (int *)&quiet,       OPT_BOOL},
      {"new",        0, (int *)&new_asn1,    OPT_BOOL},
      {"pwstdin",    0, (int *)&pwstdin,     OPT_BOOL},
      {"pin",        1, (int *)&pin,         OPT_STRING},
      {"conf",       1, NULL,                OPT_CONFIG},
      {"confile",    1, (int *)&conffile,    OPT_STRING},
      {"userconf",   1, (int *)&userconf,    OPT_STRING},
      {"voms",       1, (int *)&voms,        OPT_MULTI},
      {"order",      1, (int *)&order,       OPT_MULTI},
      {"print",      0, (int *)&print,       OPT_BOOL},
      {"globus",     1, &version,            OPT_NUM},
      {"vo",         1, (int *)&voID,        OPT_STRING},
      {"novo",       0, (int *)&novo,        OPT_BOOL},
      {"noregen",    0, (int *)&noregen,     OPT_BOOL},
      {"separate",   1, (int *)&separate,    OPT_STRING},
      {"ignorewarn", 0, (int *)&ignorewarn,  OPT_BOOL},
      {"failonwarn", 0, (int *)&failonwarn,  OPT_BOOL},
#ifdef CLASS_ADD
      {"classadd", 1, (int *)class_add_buf, OPT_STRING},
#endif
      {0, 0, 0, 0}
    };

    if (!getopts(argc, argv, opts))
      exit(1);

    if (failonwarn && ignorewarn)
      ignorewarn = false;

    if (progversion) {
      std::cout << SUBPACKAGE << "\nVersion: " << VERSION << std::endl;
      std::cout << "Compiled: " << __DATE__ << " " << __TIME__ << std::endl;
      exit(0);
    }

    if (voID=="")
      novo=true;

    version = globus(version);
    if (version == 0) {
      version = 22;
      if (debug) {
	std::cerr << "Unable to discover Globus Version: Trying for 2.2" << std::endl << std::flush;
      }
    }

    if (debug)
      std::cout << "Detected Globus Version: " << version << std::endl;

    if ((bits!=512) && (bits!=1024) && (bits!=2048) && (bits!=4096)) {
      std::cerr << "bits must be one of 512,1024,2048,4096" << std::endl;
      exit(1);
    }

    if (vomslife == -1)
      vomslife = hours;

    if (new_asn1)
      fix_add_entry_asn1_set_param = 0;
    if (pwstdin)
      pw_cb = (int (*)())(pwstdin_callback);

    (void)loadfile(conffile,0,0, debug);

    bool wasempty = false;
    if (userconf.empty()) {
      wasempty = true;
      char *uc = getenv("VOMS_USERCONF");
      if (uc)
	userconf = uc;
    }

    if (userconf.empty()) {
      char *uc = getenv("HOME");
      if (uc)
	userconf = std::string(uc) + "/" + USERCONFILENAME;
      else
	userconf = std::string("~/") + USERCONFILENAME;
    }

    if (!userconf.empty())
      (void)loadfile(userconf, getuid(), getgid(), debug, wasempty);

//    if (setrlimit(RLIMIT_CORE, &newlimit) != 0)
//      exit(1);

    exit(make_proxy() == true ? 0 : 1);
}


/*
 * Function:
 *   make_proxy()
 *
 * Description:
 *   This function is the crux of the program.  It first creates a standard
 *   certificate, than, if necessary, uses it to contact the VOMS servers,
 *   retrieves the data and puts it into an extension.  Then, it creates a
 *   new proxy with the extension. If the VOPROXY extension was enabled, it
 *   also adds the /D=vo field to the subject at this point.
 *
 * Parameters:
 *   NONE.
 *
 * Result:
 *   Failure:
 *     Non zero.
 *   Success:
 *     0.
 *
 * NOTE: Even in case of failure, there may have been enough time to complete
 * the first stage (creating a standard certificate), and this is not deleted.
 */
static bool
make_proxy()
{
  bool                   ret_status  = true;
//  proxy_cred_desc *     pcd         = NULL;
  globus_gsi_cred_handle_t issuer = NULL;
//  proxy_verify_desc     pvd;
//  proxy_verify_ctx_desc pvxd;
  BIO *                 bio_err;
  time_t                time_after;
  time_t                time_now;
  time_t                time_diff;
  time_t		time_after_proxy;
  ASN1_UTCTIME *        asn1_time = NULL;
  char *filedata = (char *)MAP_FAILED;
  std::string file;
  int   filelen = 0;
  int   fd = -1;
  std::string data;
  char *ccaf;
  char *cd;
  char *of;
  char *cf;
  char *kf;
  std::string dummy = "";
  std::string tmpproxyname = "";
  std::string uid;
  char *oldvalue=NULL;

#ifdef WIN32
  CRYPTO_malloc_init();
#endif

  globus_module_activate(GLOBUS_GSI_PROXY_MODULE); 
  globus_module_activate(GLOBUS_GSI_CALLBACK_MODULE);
  globus_module_activate(GLOBUS_GSI_SYSCONFIG_MODULE);
  globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE);
  globus_module_activate(GLOBUS_GSI_CERT_UTILS_MODULE);
  globus_module_activate(GLOBUS_GSI_GSSAPI_MODULE);
//  ERR_load_prxyerr_strings(0);
  SSLeay_add_ssl_algorithms();

  EVP_set_pw_prompt("Enter GRID pass phrase:");

  if (1) {
    if ((bio_err=BIO_new(BIO_s_file())) != NULL)
      BIO_set_fp(bio_err,stderr,BIO_NOCLOSE);

//!!!    if ((pcd = proxy_cred_desc_new()) == NULL)
//!!!      goto err;

//!!!    pcd->type = CRED_TYPE_PERMANENT;

    if(globus_gsi_cred_handle_init (&issuer, NULL) != GLOBUS_SUCCESS) 
        goto err;

    /*
     * These 5 const_cast are allowed because proxy_get_filenames will
     * overwrite the pointers, not the data itself.
     */
    ccaf = (certcafile.empty() ? NULL : const_cast<char *>(certcafile.c_str()));
    cd   = (certdir.empty()    ? NULL : const_cast<char *>(certdir.c_str()));
    of   = (outfile.empty()    ? NULL : const_cast<char *>(outfile.c_str()));
    cf   = (certfile.empty()   ? NULL : const_cast<char *>(certfile.c_str()));
    kf   = (keyfile.empty()    ? NULL : const_cast<char *>(keyfile.c_str()));
    
    if (noregen) {
      bool modify = false;
      of = NULL;
      if (cf == NULL && kf == NULL) 
	modify = true;

      if(!ccaf) ccaf=getenv("X509_CERT_FILE");
      if(!cd) if(GLOBUS_GSI_SYSCONFIG_GET_CERT_DIR(&cd) 
                                     != GLOBUS_SUCCESS) goto err;
      if(!of) if(GLOBUS_GSI_SYSCONFIG_GET_PROXY_FILENAME(&of,
                     GLOBUS_PROXY_FILE_OUTPUT) != GLOBUS_SUCCESS) goto err;
      if(!cf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(&cf,NULL)
                                     != GLOBUS_SUCCESS) goto err;
      if(!kf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(NULL,&kf)
                                     != GLOBUS_SUCCESS) goto err;

//      if ( proxy_get_filenames(pcd,0, &ccaf, &cd, &of, &cf, &kf))
//	goto err;

      if (modify)
	cf = kf = of;
      of   = (outfile.empty()    ? NULL : const_cast<char *>(outfile.c_str()));

      if(!ccaf) ccaf=getenv("X509_CERT_FILE");
      if(!cd) if(GLOBUS_GSI_SYSCONFIG_GET_CERT_DIR(&cd) 
                                     != GLOBUS_SUCCESS) goto err;
      if(!of) if(GLOBUS_GSI_SYSCONFIG_GET_PROXY_FILENAME(&of,
                     GLOBUS_PROXY_FILE_OUTPUT) != GLOBUS_SUCCESS) goto err;
      if(!cf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(&cf,NULL)
                                     != GLOBUS_SUCCESS) goto err;
      if(!kf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(NULL,&kf)
                                     != GLOBUS_SUCCESS) goto err;
//      if ( proxy_get_filenames(pcd,0, &ccaf, &cd, &of, &cf, &kf))
//	goto err;
    }
    else {
      if(!ccaf) ccaf=getenv("X509_CERT_FILE");
      if(!cd) if(GLOBUS_GSI_SYSCONFIG_GET_CERT_DIR(&cd) 
                                     != GLOBUS_SUCCESS) goto err;
      if(!of) if(GLOBUS_GSI_SYSCONFIG_GET_PROXY_FILENAME(&of,
                     GLOBUS_PROXY_FILE_OUTPUT) != GLOBUS_SUCCESS) goto err;
      if(!cf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(&cf,NULL)
                                     != GLOBUS_SUCCESS) goto err;
      if(!kf) if(GLOBUS_GSI_SYSCONFIG_GET_USER_CERT_FILENAME(NULL,&kf)
                                     != GLOBUS_SUCCESS) goto err;
//      if ( proxy_get_filenames(pcd,0, &ccaf, &cd, &of, &cf, &kf))
//        goto err;
    };


    if (noregen)
      tmpproxyname=std::string(of);
    else {
      tmpproxyname="/tmp/tmp_x509up_u" + stringify(getuid(),uid);
      tmpproxyname+="_" + stringify(getpid(), uid);
    }

    char *tmpvalue=getenv("X509_USER_PROXY");

    if (tmpvalue)
      oldvalue=strdup(tmpvalue);

    setenv("X509_USER_PROXY",tmpproxyname.c_str(),1);

    if (debug) {
      std::cout << "Files being used:" << std::endl;
      std::cout << "    cert_file: " << (ccaf ? ccaf : "none") << std::endl;
      std::cout << "    cert_dir : " << (cd ? cd : "none") << std::endl;
      std::cout << "    proxy    : " << (of ? of : "none") << std::endl;
      std::cout << "    user_cert: " << (cf ? cf : "none") << std::endl;
      std::cout << "    user_key : " << (kf ? kf : "none") << std::endl;
      std::cout << "    tmp_proxy: " << tmpproxyname << std::endl << std::flush;
    }

//    if (!cd)
//      pcd->certdir = strdup(cd);

//    if (!strncmp(cf, "SC:", 3))
//      EVP_set_pw_prompt("Enter card pin:");
//    else
      EVP_set_pw_prompt(const_cast<char *>(quiet ? "Enter GRID pass phrase:" :
					   "Enter GRID pass phrase for this identity:"));

 
    if(globus_gsi_cred_read_cert(issuer,cf) != GLOBUS_SUCCESS) goto err;

    if (!quiet) {
      char *s = NULL;

      if(globus_gsi_cred_get_subject_name(issuer,&s) != GLOBUS_SUCCESS) 
        goto err;
      std::cout << "Your identity: " << s << std::endl << std::flush;
      free(s);
    }

//    if (!strncmp(kf, "SC:", 3))
//      EVP_set_pw_prompt("Enter card pin:");

    if(globus_gsi_cred_read_key(issuer,kf,pw_cb) != GLOBUS_SUCCESS) goto err;

//    if (strncmp(cf, "SC:", 3) && !strcmp(cf,kf)) {
//      if (pcd->cert_chain == NULL)
//	pcd->cert_chain = sk_X509_new_null();
//      if (proxy_load_user_proxy(pcd->cert_chain, cf, NULL) < 0)
//	goto err;
//    } 
      
    if (debug)
      std::cout << "Output to " << of << std::endl << std::flush;

    /*
     * verify if the cert is good, i.e. is signed by one of the
     * trusted CAs.
     */
    if (verify) {
      globus_gsi_callback_data_t callback_data = NULL;
      if(globus_gsi_callback_data_init(&callback_data) != GLOBUS_SUCCESS)
        goto err;
      if(globus_gsi_callback_set_cert_dir(callback_data,cd) != GLOBUS_SUCCESS) {
        globus_gsi_callback_data_destroy(callback_data);
        goto err;
      };
      if(globus_gsi_cred_verify_cert_chain(issuer,callback_data)
                                    != GLOBUS_SUCCESS) {
	std::cerr << "verify OK" << std::endl << std::flush;
      } else {
	std::cerr << "verify failed" << std::endl << std::flush;
        globus_gsi_callback_data_destroy(callback_data);
	goto err;
      }
      globus_gsi_callback_data_destroy(callback_data);
    }
      
    asn1_time = ASN1_UTCTIME_new();
//    X509_gmtime_adj(asn1_time,0);
//    time_now = ASN1_UTCTIME_mktime(asn1_time);
    time_now = time(NULL);
    ASN1_UTCTIME_set(asn1_time,time_now);
    

    if (!noregen) {
      if (!quiet)
	std::cout << "Creating temporary proxy " << std::flush;

      if (pclocal(issuer, tmpproxyname.c_str(), hours, bits, 
                  limit_proxy, independent_proxy, old_proxy,
		  &kpcallback,
		  dummy,
		  NULL, 0,
#ifdef CLASS_ADD
		  (class_add_buf.empty() ? NULL : class_add_buf.c_str()),
		  class_add_buf.size(),
#else
		  NULL, 0
#endif
		  ))
	goto err;
	
      if (!quiet)
	std::cout << " Done" << std::endl << std::flush;
    }
  }

  for (unsigned int i = 0; i < voms.size(); i++) {
    std::string voname = voms[i];
    std::string::size_type pos = voname.find(':');
    std::string groupname = "A";
    std::string duration = "";
    std::string ordering = ",";

    for (unsigned int j =0; j < order.size(); j++)
      ordering += std::string(order[j]) + ",";

    if (pos != std::string::npos) {
      if (voname.size() > pos)
	groupname = voname.substr(pos+1);
      if (pos > 0)
	voname = voname.substr(0, pos);
    }

    bool res;
    std::string buffer;
    std::string errmsg;
    std::vector<vomsdata> servers;
    std::string contact;

    servers = findinfile(voname);

    if (!servers.empty()) {
      random_shuffle(servers.begin(), servers.end());

      for (std::vector<vomsdata>::iterator beg=servers.begin();
	   beg != servers.end();
	   beg++)
	if (debug)
	  std::cout << beg->nick << " (" << beg->host << ":" << beg->port
		    << ") [" << beg->contact << "] \"" << beg->vo << "\""
		    << std::endl << std::flush;
      
      std::vector<vomsdata>::iterator beg = servers.begin();
	   
      do {
	if (debug)
	  std::cout << "Trying: " << beg->host << ":" << beg->port
		    << ") \"" << beg->contact << "\" [" << beg->vo
		    << "]" << std::endl << std::flush;

	buffer = errmsg = "";
	res = net_connect(beg->host, beg->port, beg->contact, beg->vo,
			  stringify(vomslife*3600, duration) + ordering + groupname,
			  buffer, version, i, errmsg, quiet);
	contact = beg->contact;
	beg++;

	if (!ignorewarn) {
	  if (!errmsg.empty()) {
	    if (!quiet)
	      std::cerr << errmsg << std::flush;

	    if (failonwarn) {
	      if (!quiet)
		std::cout << "Error in getting data from VOMS server:" << contact
			  << " (or in memorizing)" << std::endl << std::flush;
	      res = false;
	    }
	  }
	}
      } while (!res && beg != servers.end());

	
      if (!res)
	goto err;
      else
	data += buffer;
    }
    else {
      if (!quiet)
	std::cout << "VOMS Server for " << voname
		  << " not known." << std::endl << std::flush;
      goto err;
    }
  }

  discardfile();

  if (!oldvalue)
    unsetenv("X509_USER_PROXY");
  else {
    setenv("X509_USER_PROXY",oldvalue,1);
    free(oldvalue);
  }

  if (!noregen)
    unlink(tmpproxyname.c_str());

  if (!incfile.empty()) {
    errno = 0;
    if ((fd = open(incfile.c_str(), O_RDONLY)) != -1) {
      filelen  = lseek(fd,0, SEEK_END);
      if (filelen != (off_t)-1)
	filedata = (char *)mmap(0, filelen, PROT_READ, MAP_PRIVATE, fd ,0);
      if (filedata == MAP_FAILED) {
	if (!quiet)
	  std::cout << "File " << incfile << " not readable!"
		    << std::endl << std::flush;
	goto err;
      }
    }
  }

  if (!separate.empty()) {
    int fd = open(separate.c_str(), O_WRONLY|O_CREAT|O_TRUNC);
    if (fd >= 0) {
      int remaining = data.size();
      int offset = 0;
      int res;
      do {
	res = write(fd, data.c_str() + offset, remaining);
	if (res == -1) {
	  perror(NULL);
	  close(fd);
	  return 1;
	}
	offset += res;
	remaining -= res;
      } while (res == remaining);
      close(fd);
    }
    else {
      perror(NULL);
      return 1;
    }
    return 0;
  }

  if (print) {
    std::string buffer = data;
    unsigned int i=0;
    while (i < buffer.size()) {
      if ((buffer[i] < 32 || buffer[i] >126) && buffer[i] != '\n')
	std::cout << '.';
      else
	std::cout  << buffer[i];
      i++;
    }
    std::cout << std::flush;

    if (!incfile.empty()) {
      int i = 0;
      while (i < filelen) {
	if ((filedata[i] < 32 || filedata[i] >126) && filedata[i] != '\n')
	  std::cout << '.';
	else
	  std::cout << filedata[i];
	i++;
      }
      std::cout << std::flush;
    }
    exit(0);
  }

  OBJ_create("1.3.6.1.4.1.8005.100.100.1","VOMS","Voms");
  OBJ_create("1.3.6.1.4.1.8005.100.100.2","INCFILE","IncFile");
  OBJ_create("1.3.6.1.4.1.8005.100.100.3","VO","Vo");

  if (noregen) {
    if ((bio_err=BIO_new(BIO_s_file())) != NULL)
      BIO_set_fp(bio_err,stderr,BIO_NOCLOSE);

//!!!    if ((pcd = proxy_cred_desc_new()) == NULL)
//!!!      goto err;

//!!!    pcd->type = CRED_TYPE_PERMANENT;

    if(globus_gsi_cred_handle_init (&issuer, NULL) != GLOBUS_SUCCESS) 
        goto err;

    if (debug) {
      std::cout << "Files being used:" << std::endl;
      std::cout << "    cert_file: " << (ccaf ? ccaf : "none") << std::endl;
      std::cout << "    cert_dir : " << (cd ? cd : "none") << std::endl;
      std::cout << "    proxy    : " << (of ? of : "none") << std::endl;
      std::cout << "    user_cert: " << (cf ? cf : "none") << std::endl;
      std::cout << "    user_key : " << (kf ? kf : "none") << std::endl << std::flush;
    }

//    if (!cd)
//      pcd->certdir = strdup(cd);

//    if (!strncmp(cf, "SC:", 3))
//      EVP_set_pw_prompt("Enter card pin:");
//    else
      EVP_set_pw_prompt(const_cast<char *>(quiet ? "Enter GRID pass phrase:" :
					   "Enter GRID pass phrase for this identity:"));

    if(globus_gsi_cred_read_cert(issuer,cf) != GLOBUS_SUCCESS) goto err;
//    if (proxy_load_user_cert(pcd, cf, pw_cb, NULL))
//      goto err;

    if (!quiet) {
      char *s = NULL;
      if(globus_gsi_cred_get_subject_name(issuer,&s) != GLOBUS_SUCCESS)
        goto err;
//      s = X509_NAME_oneline(X509_get_subject_name(pcd->ucert),NULL,0);
      std::cout << "Your identity: " << s << std::endl << std::flush;
      free(s);
    }

//    if (!strncmp(kf, "SC:", 3))
//      EVP_set_pw_prompt("Enter card pin:");
	
    if(globus_gsi_cred_read_key(issuer,kf,pw_cb) != GLOBUS_SUCCESS) goto err;
//    if (proxy_load_user_key(pcd, kf, pw_cb, NULL))
//      goto err;

//    if (strncmp(cf, "SC:", 3) && !strcmp(cf,kf)) {
//      if (pcd->cert_chain == NULL)
//	pcd->cert_chain = sk_X509_new_null();
//      if (proxy_load_user_proxy(pcd->cert_chain, cf, NULL) < 0)
//	goto err;
//    } 
      
    if (debug)
      std::cout << "Output to " << of << std::endl << std::flush;
    /*
     * verify if the cert is good, i.e. is signed by one of the
     * trusted CAs.
     */
    if (verify) {
      globus_gsi_callback_data_t callback_data = NULL;
      if(globus_gsi_callback_data_init(&callback_data) != GLOBUS_SUCCESS)
        goto err;
      if(globus_gsi_callback_set_cert_dir(callback_data,cd) != GLOBUS_SUCCESS) {        globus_gsi_callback_data_destroy(callback_data);
        goto err;
      };
      if(globus_gsi_cred_verify_cert_chain(issuer,callback_data)
                                    != GLOBUS_SUCCESS) {
//      proxy_verify_ctx_init(&pvxd);
//      proxy_verify_init(&pvd, &pvxd);
//      pvxd.certdir = cd;
//      if (proxy_verify_cert_chain(pcd->ucert,pcd->cert_chain,&pvd)) {
	if (!quiet)
	  std::cerr << "verify OK" << std::endl << std::flush;
      } else {
	if (!quiet)
	  std::cerr << "verify failed" << std::endl << std::flush;
        globus_gsi_callback_data_destroy(callback_data);
	goto err;
      }
      globus_gsi_callback_data_destroy(callback_data);
    }
      
    asn1_time = ASN1_UTCTIME_new();
//    X509_gmtime_adj(asn1_time,0);
//    time_now = ASN1_UTCTIME_mktime(asn1_time);
    time_now = time(NULL);
    ASN1_UTCTIME_set(asn1_time,time_now);
  }

  if (!quiet)
    std::cout << "Creating proxy " << std::flush;

  if (pclocal(issuer, of, hours, bits,
              limit_proxy, independent_proxy, old_proxy,
	      &kpcallback,
	      data,
	      filedata, filelen,
#ifdef CLASS_ADD
	      class_add_buf.c_str(), class_add_buf.size(),
#else
	      NULL, 0
#endif
	      ))
    goto err;

  if (!quiet)
    std::cout << " Done" << std::endl << std::flush;

  if (fd != -1) {
    if (filedata != MAP_FAILED)
      munmap(filedata,filelen);
    close(fd);
  }
  /*
   * test if expired, or the proxy will not be good for expected
   * length of time
   * But an expired certificate and proxy can be used to 
   * renew a certificate, so we still create the proxy! 
   */

  globus_gsi_cred_get_goodtill(issuer,&time_after);
//  time_after = ASN1_UTCTIME_mktime(X509_get_notAfter(pcd->ucert));
  time_diff = time_after - time_now ;
  if (time_diff < 0) {
    if (!quiet)
      std::cout << "Error: your certificate expired "
		<< asctime(localtime(&time_after)) << std::endl << std::flush;
    ret_status = false;
  } else
    if (hours && time_diff < hours*60*60) {
      if (!quiet)
	std::cout << "Warning: your certificate and proxy will expire "
		  << asctime(localtime(&time_after))
		  << "which is within the requested lifetime of the proxy"
		  << std::endl << std::flush;
      ret_status = true;
    }

  if (!quiet && ret_status == 0) {
    if (hours)
      time_after_proxy = time_now + hours*60*60;
    else
      time_after_proxy = time_after;

    if (!quiet)
      std::cout << "Your proxy is valid until "
		<< asctime(localtime(&time_after_proxy)) << std::endl << std::flush;
  }

  std::cout << std::endl << std::flush;
  globus_module_deactivate(GLOBUS_GSI_GSSAPI_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CERT_UTILS_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
  globus_module_deactivate(GLOBUS_GSI_SYSCONFIG_MODULE);
  globus_module_deactivate(GLOBUS_GSI_PROXY_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CALLBACK_MODULE);
  return ret_status;


err:
  {
    unsigned long l;
    char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
    const char *file;
#else
    char *file;
#endif
    const char *dat;
    int line;

    if (filedata != MAP_FAILED)
      munmap(filedata,filelen);

    if (fd != -1)
      close(fd);

    discardfile();

    /* WIN32 does not have the ERR_get_error_line_data */ 
    /* exported, so simulate it till it is fixed */
    /* in SSLeay-0.9.0 */
    
    while ( ERR_peek_error() != 0 ) {
      int i;
      ERR_STATE *es;
	    
      es = ERR_get_state();
      i=(es->bottom+1)%ERR_NUM_ERRORS;
	    
      if (es->err_data[i] == NULL)
	dat = "";
      else
	dat = es->err_data[i];
	    
      l=ERR_get_error_line(&file,&line);
      if (debug)
	std::cerr << ERR_error_string(l,buf) << ":"
	     << file << ":" << line << dat << std::endl << std::flush;
      else
	std::cerr << ERR_reason_error_string(l) << dat << std::endl
	     << "Function: " << ERR_func_error_string(l);
    }
  }
  globus_module_deactivate(GLOBUS_GSI_GSSAPI_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CERT_UTILS_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
  globus_module_deactivate(GLOBUS_GSI_SYSCONFIG_MODULE);
  globus_module_deactivate(GLOBUS_GSI_PROXY_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CALLBACK_MODULE);
  exit(3);
}

/**********************************************************************
Function: proxy_create_local()

Description:
    Creates a proxy on the local machine.
        
Parameters:

Returns:
**********************************************************************/

static int
pclocal(
    globus_gsi_cred_handle_t            issuer, // contains cert and key of user
    const char *                        outfile,
    int                                 hours,
    int                                 bits,
    int                                 limit_proxy,
    int                                 independent_proxy,
    int                                 old_proxy,
    void                                (*kpcallback)(int,int,void*),
    std::string &                       voms_buf,
    char *                              file_data,
    int                                 file_len,
    char *                              class_add_buf,
    int                                 class_add_buf_len)
{
    int                                 status = 1;
    globus_gsi_cert_utils_cert_type_t   cert_type =
        GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_IMPERSONATION_PROXY;
    STACK_OF(X509_EXTENSION) *          extensions = NULL;
    X509_EXTENSION                     *ex = NULL, *ex2 = NULL, 
                                       *ex3 = NULL, *ex4 = NULL;
    bool                                voms, classadd, file, vo;
    globus_result_t                     result  = GLOBUS_SUCCESS;
    globus_gsi_proxy_handle_attrs_t     proxy_handle_attrs = NULL;
    globus_gsi_proxy_handle_t           proxy_handle = NULL;
    globus_gsi_cred_handle_t            proxy_cred_handle = NULL;
    X509 *                              issuer_cert = NULL;
    EVP_PKEY *                          issuer_pkey = NULL;
    X509 * proxy_cert = NULL;
    EVP_MD *                            algorithm;

    vo = voms = classadd = file = false;

    if(independent_proxy) {
      cert_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_INDEPENDENT_PROXY;
    };
    if(limit_proxy) {
      if(GLOBUS_GSI_CERT_UTILS_IS_GSI_2_PROXY(cert_type)) {
        cert_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_LIMITED_PROXY;
      } else {
        cert_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_LIMITED_PROXY;
      };
    };
    if(old_proxy) {
      if(GLOBUS_GSI_CERT_UTILS_IS_LIMITED_PROXY(cert_type)) {
        cert_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_LIMITED_PROXY;
      } else {
        cert_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_PROXY;
      };
    };

    // Create proxy handle
    if((globus_gsi_proxy_handle_attrs_init(&proxy_handle_attrs)
                                                 != GLOBUS_SUCCESS) ||
       (globus_gsi_proxy_handle_attrs_set_keybits(proxy_handle_attrs, bits)
                                                 != GLOBUS_SUCCESS) ||
       (globus_gsi_proxy_handle_attrs_set_key_gen_callback(
                  proxy_handle_attrs,kpcallback) != GLOBUS_SUCCESS) ||
       (globus_gsi_proxy_handle_init(&proxy_handle, proxy_handle_attrs)
                                                 != GLOBUS_SUCCESS)) {
      std::cerr<<"Couldn't initialize the proxy handle"<<std::endl;
      goto err;
    };
    if(globus_gsi_proxy_handle_attrs_destroy(proxy_handle_attrs) 
                                                 != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't destroy the proxy handle attributes"<<std::endl;
      goto err;
    };
    proxy_handle_attrs=NULL;

    // Set various properties of handle
    if((globus_gsi_proxy_handle_set_time_valid(proxy_handle, hours*60)
                                                       != GLOBUS_SUCCESS) ||
       (globus_gsi_proxy_handle_set_type(proxy_handle, cert_type)
                                                       != GLOBUS_SUCCESS)) {
      std::cerr<<"Couldn't adjust proxy properties"<<std::endl;
      goto err;
    };

    // Create ordinary proxy
    if(globus_gsi_proxy_create_signed(proxy_handle, issuer, &proxy_cred_handle)
                                                       != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't create proxy credentials"<<std::endl;
      goto err;
    };

    // Create VOMS extensions
    if ((extensions = sk_X509_EXTENSION_new_null()) == NULL)
    {
        std::cerr<<"Couldn't create extensions"<<std::endl;
        goto err;
    };
    if (voms_buf.size())
    {
        if ((ex = proxy_extension_voms_create(voms_buf)) == NULL)
        {
            std::cerr<<"Couldn't create extension for voms"<<std::endl;
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex))
        {
            std::cerr<<"Couldn't add voms to list of extensions"<<std::endl;
            goto err;
        }
        voms = true;
    };

    if (file_data && file_len > 0)
    {
        if ((ex3 = proxy_extension_file_create((unsigned char *)file_data,
                                              file_len)) == NULL)
        {
            std::cerr<<"Couldn't create extension for file_data"<<std::endl;
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex3))
        {
            std::cerr<<"Couldn't add file data to list of extensions"<<std::endl;
            goto err;
        }
        file = true;
    }

    if (!novo)
    {
        if ((ex4 = proxy_extension_vo_create(voID)) == NULL)
        {
            std::cerr<<"Couldn't create extension for VO"<<std::endl;
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex4))
        {
            std::cerr<<"Couldn't add VO to list of extensions"<<std::endl;
            goto err;
        }
        vo = true;
    }

#ifdef CLASS_ADD
    if (class_add_buf && class_add_buf_len > 0)
    {
        if ((ex2 = proxy_extension_class_add_create((void *)class_add_buf,
                                                    class_add_buf_len)) == NULL)        {
            std::cerr<<"Couldn't create extension for classadds"<<std::endl;
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex2))
        {
            std::cerr<<"Couldn't add classadds to list of extensions"<<std::endl;
            goto err;
        }
        classadd = true;
    }
#endif

    // Obtain signed certificate
    if(globus_gsi_cred_get_cert(proxy_cred_handle,&proxy_cert)
                                              != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't obtain proxy certificate"<<std::endl;
      goto err;
    };

    //  Add extensions to certificate
    for(int ex_n = 0; ex_n < sk_X509_EXTENSION_num(extensions); ++ex_n) {
      X509_EXTENSION *ex = sk_X509_EXTENSION_value(extensions,ex_n);
      if(ex) {
        if(debug) {
          std::cerr<<std::endl<<"Adding extension "<<ex_n<<": ";
          ASN1_STRING_print_ex_fp(stderr,ex->value,ASN1_STRFLGS_ESC_2253 | ASN1_STRFLGS_ESC_CTRL | ASN1_STRFLGS_ESC_MSB);
          std::cerr<<std::endl;
        };
        if(!X509_add_ext(proxy_cert, ex, 0)) {
          std::cerr<<"Couldn't add extension to proxy"<<std::endl;
          goto err;
        };
      };
    };

    // Resign
    if(globus_gsi_cred_get_key(issuer, &issuer_pkey) != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't obtain issuer's private key"<<std::endl;
      goto err;
    };
    if(globus_gsi_proxy_handle_get_signing_algorithm(proxy_handle,
                       &algorithm) != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't obtain signng algorithm"<<std::endl;
      goto err;
    };
    if(!X509_sign(proxy_cert, issuer_pkey, algorithm)) {
      std::cerr<<"Couldn't resign certificate"<<std::endl;
      goto err;
    };
    
    // store modified certificate
    if(globus_gsi_cred_set_cert(proxy_cred_handle,proxy_cert)
                                              != GLOBUS_SUCCESS) {
      std::cerr<<"Couldn't put proxy certificate into handle"<<std::endl;
      goto err;
    };

    // Write proxy
    if(globus_gsi_cred_write_proxy(proxy_cred_handle, (char*)outfile)
                                                       != GLOBUS_SUCCESS) {
      std::cerr<<"The proxy credential could not be written to the output file"<<std::endl;
      goto err;
    };

    status = 0;

err:
    if(proxy_handle_attrs) globus_gsi_proxy_handle_attrs_destroy(proxy_handle_attrs);
    if(proxy_handle) globus_gsi_proxy_handle_destroy(proxy_handle);
    if(proxy_cred_handle) globus_gsi_cred_handle_destroy(proxy_cred_handle);
    if (extensions) {
        sk_X509_EXTENSION_pop_free(extensions, X509_EXTENSION_free);
        voms = false; classadd = false; file = false; vo = false;
    };
    if (voms) X509_EXTENSION_free(ex2);
    if (file) X509_EXTENSION_free(ex3);
    if (vo) X509_EXTENSION_free(ex4);
    if (classadd) X509_EXTENSION_free(ex);

    return status;
}

/**********************************************************************
Function: proxy_extension_voms_create()

Description:
            create a X509_EXTENSION for the voms info. 
        
Parameters:
                A buffer and length. The date is added as
                ANS1_OCTET_STD::STRING to an extension with the 
                class_add  OID.

Returns:

**********************************************************************/


/* As the above comment may suggest, this is a modified
 * proxy_extension_class_add_create() function that knows how to handle the
 * VOMS extension instead. */

static X509_EXTENSION *
proxy_extension_voms_create(std::string &buffer)
{
    X509_EXTENSION *                    ex = NULL;
    ASN1_OBJECT *                       voms_obj = NULL;
    ASN1_OCTET_STRING *                 voms_oct = NULL;
    int                                 crit = 0;

    if(!(voms_obj = OBJ_nid2obj(OBJ_txt2nid("VOMS"))))
    {
        goto err;
    }

    if(!(voms_oct = ASN1_OCTET_STRING_new()))
    {
        goto err;
    }

    voms_oct->data = (unsigned char *)buffer.c_str();
    voms_oct->length = buffer.size();

    if (!(ex = X509_EXTENSION_create_by_OBJ(NULL, voms_obj, 
                                            crit, voms_oct)))
    {
        goto err;
    }
    voms_oct = NULL;

    return ex;

err:
    if (voms_oct)
    {
        ASN1_OCTET_STRING_free(voms_oct);
    }
    
    if (voms_obj)
    {
        ASN1_OBJECT_free(voms_obj);
    }
    return NULL;
}

/**********************************************************************
Function: proxy_extension_file_create()

Description:
            create a X509_EXTENSION for the file info. 
        
Parameters:
                A buffer and length. The date is added as
                ANS1_OCTET_STD::STRING to an extension with the 
                class_add  OID.

Returns:

**********************************************************************/


/* As the above comment may suggest, this is a modified
 * proxy_extension_class_add_create() function that knows how to handle the
 * VOMS extension instead. */

static X509_EXTENSION *
proxy_extension_file_create(
    unsigned char *                     buffer,
    size_t                              length)

{
    X509_EXTENSION *                    ex = NULL;
    ASN1_OBJECT *                       file_obj = NULL;
    ASN1_OCTET_STRING *                 file_oct = NULL;
    int                                 crit = 0;

    if(!(file_obj = OBJ_nid2obj(OBJ_txt2nid("INCFILE"))))
    {
        goto err;
    }

    if(!(file_oct = ASN1_OCTET_STRING_new()))
    {
        goto err;
    }

    file_oct->data = buffer;
    file_oct->length = length;

    if (!(ex = X509_EXTENSION_create_by_OBJ(NULL, file_obj, 
                                            crit, file_oct)))
    {
        goto err;
    }
    file_oct = NULL;

    return ex;

err:
    if (file_oct)
    {
        ASN1_OCTET_STRING_free(file_oct);
    }
    
    if (file_obj)
    {
        ASN1_OBJECT_free(file_obj);
    }
    return NULL;
}

static X509_EXTENSION *
proxy_extension_vo_create(std::string &vo)
{
    X509_EXTENSION *                    ex = NULL;
    ASN1_OBJECT *                       vo_obj = NULL;
    ASN1_OCTET_STRING *                 vo_oct = NULL;
    int                                 crit = 0;

    if(!(vo_obj = OBJ_nid2obj(OBJ_txt2nid("VO"))))
    {
        goto err;
    }

    if(!(vo_oct = ASN1_OCTET_STRING_new()))
    {
        goto err;
    }

    vo_oct->data = (unsigned char *)vo.c_str();
    vo_oct->length = vo.size();

    if (!(ex = X509_EXTENSION_create_by_OBJ(NULL, vo_obj, 
                                            crit, vo_oct)))
    {
        goto err;
    }
    vo_oct = NULL;

    return ex;

err:
    if (vo_oct)
    {
        ASN1_OCTET_STRING_free(vo_oct);
    }
    
    if (vo_obj)
    {
        ASN1_OBJECT_free(vo_obj);
    }
    return NULL;
}



static X509_EXTENSION *
proxy_extension_create(std::string name, std::string &value, bool critic=false)
{
    X509_EXTENSION *                    ex = NULL;
    ASN1_OBJECT *                       obj = NULL;
    ASN1_OCTET_STRING *                 oct = NULL;
    int                                 crit = (critic ? 1 : 0);

    if(!(obj = OBJ_nid2obj(OBJ_txt2nid(const_cast<char *>(name.c_str())))))
    {
        goto err;
    }

    if(!(oct = ASN1_OCTET_STRING_new()))
    {
        goto err;
    }

    oct->data = (unsigned char *)value.c_str();
    oct->length = value.size();

    if (!(ex = X509_EXTENSION_create_by_OBJ(NULL, obj, crit, oct)))
    {
        goto err;
    }
    oct = NULL;

    return ex;

err:
    if (oct)
    {
        ASN1_OCTET_STRING_free(oct);
    }
    
    if (obj)
    {
        ASN1_OBJECT_free(obj);
    }
    return NULL;
}


