/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/* The following code declares classes to read from and write to
 * file descriptore or file handles.
 *
 * See
 *      http://www.josuttis.com/cppcode
 * for details and the latest version.
 *
 * - open:
 *      - integrating BUFSIZ on some systems?
 *      - optimized reading of multiple characters
 *      - stream for reading AND writing
 *      - i18n
 *
 * (C) Copyright Nicolai M. Josuttis 2001.
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 * Version: Jul 28, 2002
 * History:
 *  Jul 28, 2002: bugfix memcpy() => memmove()
 *                fdinbuf::underflow(): cast for return statements
 *  Aug 05, 2001: first public version
 */
/*
 * Further changes to the fdstream code, by vincenzo:
 *  May 28, 2003: removed boost references and fdistream code and
 *                added empty constructors (vincenzo)
 */

#include <iostream>
#include <fstream>

#if 0
#include "GSISocketServer.h"
#include "GSISocketAgent.h"
#else
#include "Server.h"
#endif

#include <openssl/evp.h>

#include "data.h"

#if (__GNUC__ >= 3) 
#include <ext/stdio_filebuf.h>
typedef class __gnu_cxx::stdio_filebuf<char> myfilebuf; 

#include <istream>
#include <ostream>
#include <streambuf>
// for EOF:
#include <cstdio>
// for memmove():
#include <cstring>


// low-level read and write functions
#ifdef _MSC_VER
# include <io.h>
#else
# include <unistd.h>
//extern "C" {
//    int write (int fd, const char* buf, int num);
//    int read (int fd, char* buf, int num);
//}
#endif



/************************************************************
 * fdostream
 * - a stream that writes on a file descriptor
 ************************************************************/


class fdoutbuf : public std::streambuf {
  protected:
    int fd;    // file descriptor
  public:
    // constructor
    fdoutbuf (int _fd) : fd(_fd) {
    }
    fdoutbuf () {}
    void set(int _fd) { fd = _fd; }
  protected:
    // write one character
    virtual int_type overflow (int_type c) {
        if (c != EOF) {
            char z = c;
            if (write (fd, &z, 1) != 1) {
                return EOF;
            }
        }
        return c;
    }
    // write multiple characters
    virtual
    std::streamsize xsputn (const char* s,
                            std::streamsize num) {
        return write(fd,s,num);
    }
};

class fdostream : public std::ostream {
  protected:
    fdoutbuf buf;
  public:
    fdostream (int fd) : std::ostream(0), buf(fd) {
        rdbuf(&buf);
    }
    fdostream() :std::ostream(0) {}
    void open(int fdes) { buf.set(fdes); rdbuf(&buf); }
};

#endif

class VOMSInitException {};

class VOMSServer {
public:
  VOMSServer(int argc, char *argv[]);
  ~VOMSServer();

  void Run();
private:
  //  VOMSServer(VOMSServer const &old) : sock(9999, version) {fbuf=old.fbuf;}
  VOMSServer &operator=(VOMSServer const &) {exit(1);}
  GSISocketServer sock;

  bool Execute(const std::string &, const std::string &, EVP_PKEY *);
  bool Initialize(int argc, char *argv[]);

#if (__GNUC__ >=3)
  fdostream logger;
#else
  std::ofstream logger;
#endif

  int         validity;
  std::string logfile;
  bool        gatekeeper_test;
  int         daemon_port;
  bool        foreground;
  std::string globuspwd;
  std::string globusid;
  std::string x509_cert_dir;
  std::string x509_cert_file;
  std::string x509_user_proxy;
  std::string x509_user_cert;
  std::string x509_user_key;
  std::string desired_name_char;
  std::string username;
  std::string dbname;
  std::string passfile;
  std::string voname;
  std::string uri;
  int         version;
  std::string subject;
  std::string ca;
};
