#include "../../std.h"

#include <globus_io.h>

#include "../../misc/globus_error_utils.h"
#include "../../misc/log_time.h"


#include "../../misc/inttostring.h"
#include "../../auth/auth.h"

#include "client.h"


static const char* W_names1[7] = {
  "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
};

static const char* W_names2[7] = {
  "Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"
};

static const char* M_names[12] = {
  "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};

static time_t http_time(const char* str) {
  char         W[32];
  unsigned int Y;
  char         M[32];
  unsigned int D;
  unsigned int h;
  unsigned int m;
  unsigned int s;
  unsigned int n;
  char Z[32];

  time_t t = (time_t)(-1);
  struct tm tt;
  if((str == NULL) || ((*str) == 0)) return false;
  // RFC1123
  if(sscanf(str,
            "%31s, %2u %31s %4u %2u:%2u:%2u %31s",
            W,&tt.tm_mday,M,&tt.tm_year,&tt.tm_hour,&tt.tm_min,&tt.tm_sec,Z) 
                                                                       == 8) {
    if(strcmp(Z,"GMT") != 0) return t;
    tt.tm_year-=1900;
  }
  // RFC850
  else if(sscanf(str,
          "%31s, %2u-%31s-%2u %2u:%2u:%2u %31s",
          W,&tt.tm_mday,M,&tt.tm_year,&tt.tm_hour,&tt.tm_min,&tt.tm_sec,Z)
                                                                      == 8) {
    if(strcmp(Z,"GMT") != 0) return t;
  }
  // asctime
  else if(sscanf(str,
          "%31s, %31s %4u %2u:%2u:%2u %4u",
          W,M,&tt.tm_mday,&tt.tm_hour,&tt.tm_min,&tt.tm_sec,&tt.tm_year) == 8) {
    tt.tm_year-=1900;
  }
  else { return t; };
  for(n=0;n<12;n++) if(strcmp(M_names[n],W) == 0) break;
  if(n>=12) return t;
  tt.tm_mon=n;
  t=timegm(&tt);
  return t;
}
 

HTTP_ResponseHeader::HTTP_ResponseHeader(bool alive):keep_alive(alive),content_length_passed(false),content_range_passed(false) { };

void HTTP_ResponseHeader::reset(bool alive) {
  keep_alive=alive;
  content_length_passed=false;
  content_range_passed=false;
}

bool HTTP_ResponseHeader::set(const char* name,const char* value) {
  if(strcasecmp("Connection:",name) == 0) {
    if(strcasecmp("close",value) == 0) { keep_alive=0; }
    else if(strcasecmp("keep-alive",value) == 0) { keep_alive=1; }
    else { return false; };
    return true;
  } else if(strcasecmp("Content-Length:",name) == 0) {
    content_length_passed=false;
    char *e;
    content_length=strtoull(value,&e,10);
    if((*e) != 0) return false;
    content_length_passed=true;
    return true;
  } else if(strcasecmp("Content-Range:",name) == 0) {
    content_range_passed=false;
    content_size=0;
    const char* p = value; for(;*p;p++) if(isspace(*p)) break;
    if(strncasecmp("bytes",value,p-value) != 0) return false;
    for(;*p;p++) if(!isspace(*p)) break;
    char *e;
    content_start=strtoull(p,&e,10);
    if((*e) != '-') return false; 
    p=e+1; content_end=strtoull(p,&e,10); p=e;
    if(((*e) != '/') && ((*e) != 0)) return false;
    if(content_start > content_end) return false;
    if((*p) == '/') {
      p++; content_size=strtoull(p,&e,10);
      if((*e) != 0) { return false; }
    };
    content_range_passed=true;
    return true;
  } else if(strcasecmp("Expires:",name) == 0) {
    expires=http_time(value);
    expires_passed=(expires != (time_t)(-1));
  } else if(strcasecmp("Last-Modified:",name) == 0) {
    last_modified=http_time(value);
    last_modified_passed=(last_modified != (time_t)(-1));
  };
  return true;
}

